from aio_client.provider.api import PostProviderRequest
from aio_client.provider.api import push_request

from concentrator.exceptions import ValidationError
from concentrator.smev3_v321.application_admission_request.constants import (
    APPLICATION_ADMISSION_REQUEST_ERROR_COMMENT)
from concentrator.smev3_v321.application_admission_request.constants import (
    DIRECTS_NOT_FOUND_ERROR_COMMENT)
from concentrator.smev3_v321.application_admission_request.constants import (
    MULTIPLE_DIRECTS_FOUND_ERROR_COMMENT)
from concentrator.smev3_v321.application_admission_request.constants import (
    SUCCESS_CODE)
from concentrator.smev3_v321.constants import ERROR_ORG_CODE
from kinder.core.children.models import ChildrenDelegate
from kinder.core.children.models import Delegate
from kinder.core.children.models import DelegateTypeEnumerate
from kinder.core.children.models import GenderEnumerate
from kinder.core.declaration.enum import DeclarationSourceEnum
from kinder.core.declaration.enum import DeclarationTypeInteractionEnum
from kinder.core.declaration.models import Declaration
from kinder.core.dict.models import DULDelegateType
from kinder.core.dict.models import GroupTypeEnumerate
from kinder.core.dict.models import GroupOrientationDocuments
from kinder.core.dict.models import GroupSpec
from kinder.core.dict.models import GroupType
from kinder.core.dict.models import HealthNeed
from kinder.core.dict.models import WorkType
from kinder.core.direct.models import DirectDeclarationEnrollment
from kinder.core.direct.models import DirectDeclarationSourceEnum
from kinder.core.direct.models import DRS
from kinder.core.direct.report import DeclarationReporter

from concentrator.smev3_v321.base import constants as base_enums
from concentrator.smev3_v321.base.exceptions import ContentFailure
from concentrator.smev3_v321.base.utils import BaseExecutor
from concentrator.smev3_v321.base.utils import SMEV3Response
from concentrator.smev3_v321.base.utils import render_type2xml
from concentrator.smev3_v321.service_types import kinder_conc
from concentrator.smev3_v321.models import ApplicantAnswer

from ..helpers import get_declaration_by_client_or_portal_id


class ApplicationAdmissionRequestExecutor(BaseExecutor):
    """Исполнитель сервиса GetApplicationAdmissionRequest."""

    name_service = 'ApplicationAdmissionRequest'
    type_service = kinder_conc.ApplicationAdmissionRequestType

    @classmethod
    def prepare_query(cls):
        """Выполняет подготовку запроса."""

        return Declaration.objects.prefetch_related(
            'children__childrendelegate_set')

    @classmethod
    def get_direct(cls, declaration):
        register_directs = declaration.direct_set.filter(
            status__code=DRS.REGISTER)
        if not register_directs.exists():
            raise ContentFailure(
                ERROR_ORG_CODE, DIRECTS_NOT_FOUND_ERROR_COMMENT)
        elif register_directs.count() > 1:
            raise ContentFailure(
                ERROR_ORG_CODE, MULTIPLE_DIRECTS_FOUND_ERROR_COMMENT)
        return register_directs.first()

    @classmethod
    def process(cls, message, request_body, **kwargs):

        declaration, _ = get_declaration_by_client_or_portal_id(
            cls.prepare_query(),
            request_body.ApplicationAdmissionRequest.orderId)

        try:
            response_body = render_type2xml(
                cls.get_response(
                    declaration, request_body.ApplicationAdmissionRequest),
                name_type='FormDataResponse')
            content_failure_comment = None
        except ContentFailure as exc:
            response_body = 'null'
            content_failure_comment = exc.content_failure_comment

        response = push_request(PostProviderRequest(
            origin_message_id=message['origin_message_id'],
            body=response_body,
            message_type=message['message_type'],
            replay_to=message['replay_to'],
            is_test_message=message['is_test_message']))

        return SMEV3Response(
            response,
            {
                'method_name': cls.name_service,
                'response': response_body,
                'result': content_failure_comment or base_enums.SUCCESS_MESSAGE
            })

    @classmethod
    def get_response(cls, declaration, request):
        """Формирует ответ на GetApplicationAdmissionRequest.

        :param declaration: Заявление
        :type declaration: Declaration
        :param request: Запрос
        :type request:
        :return: ответ
        :rtype: FormDataResponseType
        :raise: ContentFailure
        """

        if not declaration:
            raise ContentFailure(
                ERROR_ORG_CODE, APPLICATION_ADMISSION_REQUEST_ERROR_COMMENT)

        direct = cls.get_direct(declaration)

        child = declaration.children
        children_delegate_exists = child.childrendelegate_set.exists()

        if children_delegate_exists:
            children_delegate = child.childrendelegate_set.order_by('id')[0]
        else:
            children_delegate = None

        person_info = request.PersonInfo

        patronymic = person_info.PersonMiddleName
        person_doc_info = request.PersonIdentityDocInfo

        delegate_params = {
            'surname': person_info.PersonSurname,
            'firstname': person_info.PersonName,
            'patronymic': patronymic,
            'phones': person_info.PersonPhone,
            'email': person_info.PersonEmail,
            'dul_type': DULDelegateType.objects.filter(
                esnsi_code=person_doc_info.IdentityDocName.code).first(),
            'dul_series': person_doc_info.IdentityDocSeries,
            'dul_number': person_doc_info.IdentityDocNumber,
            'dul_date': person_doc_info.IdentityDocIssueDate,
            'dul_place': person_doc_info.IdentityDocIssueCode,
            'dul_issued_by': person_doc_info.IdentityDocIssued,
        }

        delegate_params = update_middlename_params(
            patronymic, delegate_params)

        if not children_delegate:
            delegate = Delegate.objects.create(**delegate_params)
            ChildrenDelegate.objects.create(children=child, delegate=delegate)
        else:
            delegate = children_delegate.delegate
            for param_name, value in delegate_params.items():
                setattr(delegate, param_name, value)
            delegate.save()

        person2info = request.Person2Info

        if person2info:
            person2middlename = person2info.Person2MiddleName

            delegate2_params = {
                'surname': person2info.Person2Surname,
                'firstname': person2info.Person2Name,
                'patronymic': person2middlename,
                'phones': person2info.Person2Phone,
                'email': person2info.Person2Email,
            }

            delegate2_params = update_middlename_params(
                person2middlename, delegate2_params)

            # Если совпали типы у родителей, то определяем второго как
            # законного представителя
            if delegate2_params.get('type') == delegate_params.get('type'):
                delegate2_params['type'] = DelegateTypeEnumerate.LEX

            if child.childrendelegate_set.count() > 1:
                child_delegate2 = child.childrendelegate_set.order_by('id')[1]
                delegate2 = Delegate.objects.filter(
                    id=child_delegate2.delegate.id).first()
                if delegate2:
                    for param_name, value in delegate2_params.items():
                        setattr(delegate2, param_name, value)
                    delegate2.save()
            else:
                delegate2 = Delegate.objects.create(**delegate2_params)
                ChildrenDelegate.objects.create(
                    children=child, delegate=delegate2)

        child_info = request.ChildInfo
        child.surname = child_info.ChildSurname
        child.firstname = child_info.ChildName
        child.patronymic = child_info.ChildMiddleName
        child.date_of_birth = child_info.ChildBirthDate
        doc_info = child_info.ChildBirthDocRF
        child.dul_series = doc_info.ChildBirthDocSeries
        child.dul_number = doc_info.ChildBirthDocNumber
        child.dul_date = doc_info.ChildBirthDocIssueDate
        child.zags_act_number = doc_info.ChildBirthDocActNumber
        child.zags_act_date = doc_info.ChildBirthDocActDate
        child.zags_act_place = doc_info.ChildBirthDocIssued

        address = request.Address
        child.address_full = address.FullAddress
        child.address_place = address.Region.code
        child.address_street = address.Street.code
        child.address_house_guid = address.House.code
        child.address_corps = address.Building1
        child.address_flat = address.Apartment

        entry_params = request.EntryParams
        declaration.spec = GroupSpec.objects.filter(
            esnsi_code=entry_params.Language.code).first()
        declaration.desired_date = entry_params.EntryDate
        declaration.work_type = WorkType.objects.filter(
            esnsi_code=entry_params.Schedule.code).first()
        agreement_on_full_day = entry_params.AgreementOnFullDayGroup

        if agreement_on_full_day:
            declaration.consent_full_time_group = agreement_on_full_day

        adaptation_program = request.AdaptationProgram
        desired_group_type = GroupType.objects.filter(
            esnsi_code=adaptation_program.AdaptationGroup.code).first()
        declaration.desired_group_type = desired_group_type

        health_need_query = HealthNeed.objects.filter(
            group_type=desired_group_type)
        # AdaptationGroupType необязательный, если группа общеразвивающая
        if desired_group_type.code != GroupTypeEnumerate.DEV:
            health_need_query = health_need_query.filter(
                esnsi_code=adaptation_program.AdaptationGroupType.code)
        child.health_need = health_need_query.first()

        agreement_on_general_group = adaptation_program.AgreementOnGeneralGroup

        if agreement_on_general_group:
            declaration.consent_dev_group = agreement_on_general_group

        agreement_on_care = adaptation_program.AgreementOnCareGroup

        if agreement_on_care:
            declaration.consent_care_group = agreement_on_care

        need_special_care = adaptation_program.NeedSpecialCareConditions

        if need_special_care:
            child.health_need_special_support = need_special_care

        medical_report = request.MedicalReport

        if medical_report:
            child.health_number = medical_report.DocNumber
            child.health_need_start_date = medical_report.DocIssueDate
            child.health_issued_by = medical_report.DocIssued
            child.health_need_expiration_date = (
                medical_report.DocExpirationDate)
            child.health_series = medical_report.DocSeries

            try:
                group_orientation_document = (
                    GroupOrientationDocuments.objects.get(
                        esnsi_code=medical_report.DocName.code))
            except GroupOrientationDocuments.DoesNotExist:
                raise ValidationError('Указано неверное значение DocName')

            child.health_need_confirmation = group_orientation_document

        child.save()
        declaration.save()
        params = dict(
            direct_id=direct.id, delegate_id=delegate.id,
            doc_list_review=request.DocListReview,
            license_charter=request.LicenseCharter,
        )
        rep = DeclarationReporter(params, params)
        report_url = rep.make_report()
        DirectDeclarationEnrollment.objects.create(
            service_type=request.ServicesType,
            direct=direct,
            user=None,
            file=report_url,
            source=DirectDeclarationSourceEnum.EPGU,
        )

        ApplicantAnswer.objects.update_or_create(
            direct=direct, defaults=dict(answer=True, comment='')
        )

        response = kinder_conc.FormDataResponseType(
            changeOrderInfo=kinder_conc.changeOrderInfoType(
                orderId=kinder_conc.orderIdType(request.orderId),
                statusCode=kinder_conc.statusCodeType(SUCCESS_CODE),
            )
        )

        return response


def update_middlename_params(middlename, params):
    """Обновляет параметры родителя, устанавливая пол при необходимости

    :param middlename: Отчество представителя
    :type middlename: str
    :param params: Словарь с параметрами представителя
    :type params: dict
    """

    middlename = middlename.strip()

    if middlename.endswith(('вна', 'кызы', 'ична')):
        params['gender'] = GenderEnumerate.FEMALE
        params['type'] = DelegateTypeEnumerate.MOTHER
    elif middlename.endswith(('ич', 'оглы')):
        params['gender'] = GenderEnumerate.MALE
        params['type'] = DelegateTypeEnumerate.FATHER

    return params
